@@ -28,6 +28,10 @@ directory as this file. Edit the already present json file as follows:
 - "GRIST_SELF_MANAGED_SINGLE_ORG": "Y" if you are running the mono-team flavour 
 - "GRIST_TEST_RUN_USER_TESTS": "N" to skip the tests that create users 
   (special, see below)
+- "GRIST_TEST_RUN_SCIM_TESTS": "N" to skip all SCIM api tests (special, see below)
+- "GRIST_TEST_RUN_EXT_ATTACH_TESTS": "N" to skip the tests that need an external 
+  storage for attachments (special, see below)
+
 Note: no other config key should be configured for the test suite. 
 
 The test suite will leave several objects (workspaces and docs) in your 
@@ -50,6 +54,21 @@ To help you keep test users under control, we added a special configuration
 key, only useful when running the test suite: set "GRIST_TEST_RUN_USER_TESTS" 
 to "N" to skip every test that may create a user. 
 
+SCIM apis are a separate problem: SCIM tests are automatically skipped when you 
+run the test suite against the regular Grist SaaS (because SCIM is not enabled 
+there); but, even in self-hosted Grist, you have to enable SCIM first by 
+setting the GRIST_ENABLE_SCIM env variable. If you are self-hosted, but you 
+don't want to enable SCIM in your environment, set our special configuration key 
+"GRIST_TEST_RUN_SCIM_TESTS" to "N" (in config_test.json file) to skip all SCIM 
+tests altogether, so you don't get any spurious errors. 
+
+Another separate problem is how to test the Apis that require an external storage 
+for attachments. If your testing environment already supports external storage 
+(that is, you have a bucket in place and a "GRIST_EXTERNAL_ATTACHMENTS_MODE" 
+variable set to "snapshots") then you may set our special configuration key 
+"GRIST_TEST_RUN_EXT_ATTACH_TESTS" to include the tests dealing with storages. 
+If not, leave the key unset (the default) to skip those tests. 
+
 A few object will also be downloaded in your current directory. Moreover, 
 tests involving user/permission manipulation may trigger email notifications 
 from Grist. 
@@ -70,14 +89,16 @@ in normal usage it's just ``<grist>.add_records(<table>, records)``.
 
 """
 
-import os, os.path
+from pathlib import Path
 import time
 from datetime import datetime
 import json
 import unittest
-from requests import HTTPError, ConnectTimeout
+from requests import HTTPError, ConnectTimeout, PreparedRequest
 
 from pygrister import api
+from pygrister import config
+from pygrister.exceptions import *
 
 # standard config values used by tests
 default_config = {
@@ -86,12 +107,12 @@ default_config = {
     "GRIST_API_ROOT": "api",
     "GRIST_RAISE_ERROR": "Y",
     "GRIST_SAFEMODE": "N",
-    "GRIST_WORKSPACE_ID": "_bogus_", # no test should ever hit this
-    "GRIST_DOC_ID": "_bogus_",       # no test should ever hit this
+    "GRIST_WORKSPACE_ID": "1000000000000", # no test should ever hit this
+    "GRIST_DOC_ID": "_bogus_",             # no test should ever hit this
 }
 
-HERE = os.path.dirname(os.path.abspath(__file__))
-with open(os.path.join(HERE, 'config_test.json')) as f:
+HERE = Path(__file__).absolute().parent
+with open(HERE / 'config_test.json') as f:
     TEST_CONFIGURATION = json.loads(f.read())
 TEST_CONFIGURATION.update(default_config)
 
@@ -105,14 +126,14 @@ def _make_ws(team_id):
     gristapi = api.GristApi(config=TEST_CONFIGURATION)
     name = 'ws'+str(time.time_ns())
     st, ws_id = gristapi.add_workspace(name, team_id)
-    total_apicalls.append(gristapi.apicalls)
+    total_apicalls.append(gristapi.apicaller.apicalls)
     return ws_id, name
 
 def _make_doc(ws_id):
     gristapi = api.GristApi(config=TEST_CONFIGURATION)
     name = 'doc'+str(time.time_ns())
     st, doc_id = gristapi.add_doc(name, ws_id=ws_id)
-    total_apicalls.append(gristapi.apicalls)
+    total_apicalls.append(gristapi.apicaller.apicalls)
     return doc_id
 
 
@@ -121,20 +142,71 @@ class BaseTestPyGrister(unittest.TestCase):
         self.g = api.GristApi(config=TEST_CONFIGURATION)
     
     def tearDown(self):
-        total_apicalls.append(self.g.apicalls)
+        total_apicalls.append(self.g.apicaller.apicalls)
 
 
 class TestVarious(BaseTestPyGrister):
     @classmethod
     def setUpClass(cls):
         cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']
+        cls.ws_id, name = _make_ws(cls.team_id)
+        cls.doc_id = _make_doc(cls.ws_id)
 
-    def test_raise_error(self):
+    def test_raise_error_before_call(self):
+        # an example where Pygrister will fail even before posting the call
+        with self.assertRaises(OSError):
+            self.g.upload_attachments(['bogus_upload_file'], 
+                                      doc_id=self.doc_id, team_id=self.team_id)
+        self.assertIsNone(self.g.apicaller.request)
+        self.assertFalse(self.g.ok) # this is by design, see note in Apicall.ok
+
+    def test_raise_error_no_response(self):
+        # an example where Grist won't return the call
+        # (this is the same scenario as test_request_options)
+        self.g.update_config({'GRIST_SERVER_PROTOCOL': 'http://', 
+                              'GRIST_TEAM_SITE': '10', 
+                              'GRIST_API_SERVER': '255.255.1', 
+                              # if we are self-managed instead
+                              'GRIST_SELF_MANAGED_HOME': 'http://10.255.255.1'})
+        self.g.apicaller.request_options = {'timeout': 1}
+        with self.assertRaises(ConnectTimeout):
+            self.g.see_team()
+        self.assertIsNone(self.g.apicaller.response)
+        self.assertFalse(self.g.apicaller.ok) # this is by design, see note in Apicall.ok
+
+    def test_raise_error_http401(self):
+        # an example where Grist won't return regular json: see that we don't crash 
+        self.g.update_config({'GRIST_API_KEY': 'bogus_api_key'})
         with self.assertRaises(HTTPError):
-            self.g.list_workspaces('_bogus_team_id_')
+            self.g.see_team()
+        self.assertEqual(self.g.apicaller.response.status_code, 401)
         self.g.update_config({'GRIST_RAISE_ERROR': 'N'})
-        st, res = self.g.list_workspaces('_bogus_team_id_')
-        self.assertGreaterEqual(st, 300)
+        st, res = self.g.see_team()
+        self.assertEqual(st, 401)
+    
+    def test_raise_error_http404(self):
+        # an example where Grist will return a more polite json object
+        with self.assertRaises(HTTPError):
+            self.g.see_team('bogus_team_id')
+        self.g.update_config({'GRIST_RAISE_ERROR': 'N'})
+        st, res = self.g.see_team('bogus_team_id')
+        self.assertEqual(st, 404)
+    
+    def test_dry_run(self):
+        self.g.apicaller.dry_run = True
+        st, res = self.g.see_team()
+        self.assertEqual(st, 418)
+        self.assertFalse(self.g.ok)
+        self.assertIsNone(self.g.apicaller.response)
+    
+    def test_safemode(self):
+        self.g.update_config({'GRIST_SAFEMODE': 'Y'})
+        with self.assertRaises(GristApiInSafeMode):
+            self.g.add_workspace('bogus')
+        # safe mode leverages on dry run but should clean up after
+        self.assertFalse(self.g.apicaller.dry_run)
+        self.assertIsNone(self.g.apicaller.response)
+        self.assertIsInstance(self.g.apicaller.request, PreparedRequest)
 
     def test_ok(self):
         try:
@@ -145,20 +217,33 @@ class TestVarious(BaseTestPyGrister):
         st, res = self.g.see_team()
         self.assertTrue(self.g.ok)
 
+    def test_custom_configurator(self):
+        self.assertEqual(self.g.configurator.config['GRIST_RAISE_ERROR'], 'Y')
+        c = config.Configurator(config={'GRIST_RAISE_ERROR': 'N'})
+        old_c = self.g.configurator
+        self.g.configurator = c
+        self.assertEqual(self.g.configurator.config['GRIST_RAISE_ERROR'], 'N')
+        self.g.configurator = old_c
+        self.assertEqual(self.g.configurator.config['GRIST_RAISE_ERROR'], 'Y')
+        g = api.GristApi(custom_configurator=c)
+        self.assertEqual(g.configurator.config['GRIST_RAISE_ERROR'], 'N')
+        with self.assertRaises(GristApiNotConfigured):
+            g = api.GristApi(config={'foo':'bar'}, custom_configurator=c)
+
     def test_reconfig(self):
-        self.assertEqual(self.g._config['GRIST_RAISE_ERROR'], 'Y')
-        self.assertEqual(self.g._config['GRIST_SAFEMODE'], 'N')
+        self.assertEqual(self.g.configurator.config['GRIST_RAISE_ERROR'], 'Y')
+        self.assertEqual(self.g.configurator.config['GRIST_SAFEMODE'], 'N')
         # update_config is for incremental changes
         self.g.update_config({'GRIST_SAFEMODE': 'Y'})
-        self.assertEqual(self.g._config['GRIST_RAISE_ERROR'], 'Y')
-        self.assertEqual(self.g._config['GRIST_SAFEMODE'], 'Y')
+        self.assertEqual(self.g.configurator.config['GRIST_RAISE_ERROR'], 'Y')
+        self.assertEqual(self.g.configurator.config['GRIST_SAFEMODE'], 'Y')
         self.g.update_config({'GRIST_RAISE_ERROR': 'N'})
-        self.assertEqual(self.g._config['GRIST_RAISE_ERROR'], 'N')
-        self.assertEqual(self.g._config['GRIST_SAFEMODE'], 'Y')
+        self.assertEqual(self.g.configurator.config['GRIST_RAISE_ERROR'], 'N')
+        self.assertEqual(self.g.configurator.config['GRIST_SAFEMODE'], 'Y')
         # reconfig is for re-building from scratch
         self.g.reconfig({'GRIST_RAISE_ERROR': 'Y'})
-        self.assertEqual(self.g._config['GRIST_RAISE_ERROR'], 'Y')
-        self.assertEqual(self.g._config['GRIST_SAFEMODE'], 'N')
+        self.assertEqual(self.g.configurator.config['GRIST_RAISE_ERROR'], 'Y')
+        self.assertEqual(self.g.configurator.config['GRIST_SAFEMODE'], 'N')
 
     def test_request_options(self):
         # as an example of extra-options, we test a timeout limit
@@ -169,7 +254,7 @@ class TestVarious(BaseTestPyGrister):
                               # maybe we are self-managed instead
                               'GRIST_SELF_MANAGED_HOME': 'http://10.255.255.1'})
         # without this, test will take forever, before failing with ConnectTimeout
-        self.g.request_options = {'timeout': 1}
+        self.g.apicaller.request_options = {'timeout': 1}
         with self.assertRaises(ConnectTimeout):
             st, res = self.g.see_team()
 
@@ -182,7 +267,7 @@ class TestVarious(BaseTestPyGrister):
         # POST
         st, ws_id = self.g.add_workspace(name, self.team_id)
         self.assertEqual(st, 200)
-        self.assertIn('Cookie', self.g.req_headers) # the session thing is working
+        self.assertIn('Cookie', self.g.apicaller.request.headers) # the session thing is working
         name = str(time.time_ns())
         st, doc_id = self.g.add_doc(name, ws_id=ws_id)
         self.assertEqual(st, 200)
@@ -191,8 +276,8 @@ class TestVarious(BaseTestPyGrister):
                                       doc_id=doc_id, team_id=self.team_id)
         self.assertEqual(st, 200)
         # POST in upload mode
-        f = os.path.join(HERE, 'imgtest.jpg')
-        st, res = self.g.upload_attachment(f, doc_id=doc_id, team_id=self.team_id)
+        f = HERE / 'imgtest.jpg'
+        st, res = self.g.upload_attachments([f], doc_id=doc_id, team_id=self.team_id)
         self.assertEqual(st, 200)
         # PATCH
         name = str(time.time_ns())
@@ -216,6 +301,7 @@ class TestUsersNoScim(BaseTestPyGrister):
 # note: to test scim ops, you need both GRIST_SELF_MANAGED here (because the 
 # regular Grist SaaS does not support scim) and GRIST_ENABLE_SCIM on the server!
 @unittest.skipIf(TEST_CONFIGURATION['GRIST_SELF_MANAGED'] == 'N', '')
+@unittest.skipIf(TEST_CONFIGURATION['GRIST_TEST_RUN_SCIM_TESTS'] == 'N', '')
 class TestUsersScim(BaseTestPyGrister):
     @classmethod
     def setUpClass(cls):
@@ -280,8 +366,6 @@ class TestUsersScim(BaseTestPyGrister):
         # now delete the user
         st, res = self.g.delete_user(user_id)
         self.assertEqual(st, 204)
-        # if the following fails, it means Grist fixed the api,  
-        # which now returns no response body at all
         self.assertIsNone(res)
         # now double-check that the user is gone
         self.g.update_config({'GRIST_RAISE_ERROR': 'N'})
@@ -366,6 +450,12 @@ class TestTeamSites(BaseTestPyGrister):
         self.assertEqual(st, 200)
         self.assertIsNone(res)
 
+    @unittest.skip
+    def test_delete_team(self):
+        # since there is no way to *create* a team in the first place, 
+        # we won't test team deletion!
+        pass
+
     def test_list_team_users(self):
         st, res = self.g.list_team_users(TestTeamSites.team_id)
         self.assertIsInstance(res, list)
@@ -388,7 +478,7 @@ class TestWorkspaces(BaseTestPyGrister):
         st, res = gristapi.list_workspaces(cls.team_id) # should work :)
         cls.workspace_id = res[0]['id']
         cls.workspace_name = res[0]['name']
-        total_apicalls.append(gristapi.apicalls)
+        total_apicalls.append(gristapi.apicaller.apicalls)
 
     def test_list_workspaces(self):
         st, res = self.g.list_workspaces(self.team_id)
@@ -471,6 +561,15 @@ class TestDocs(BaseTestPyGrister):
         with self.assertRaises(api.GristApiInSafeMode):
             self.g.add_doc(name, self.workspace_id)
 
+    def test_delete_doc_history(self):
+        name = str(time.time_ns())
+        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
+        self.assertIsInstance(doc_id, str)
+        self.assertEqual(st, 200)
+        st, res = self.g.delete_doc_history(0, doc_id)
+        self.assertIsNone(res)
+        self.assertEqual(st, 200)
+
     @unittest.expectedFailure
     def test_create_delete_doc_cross_site(self):
         # fun fact: cross-site doc creation is allowed...
@@ -519,6 +618,15 @@ class TestDocs(BaseTestPyGrister):
         self.assertIsNone(res)
         self.assertEqual(st, 200)
 
+    def test_reload_doc(self):
+        name = str(time.time_ns())
+        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
+        self.assertIsInstance(doc_id, str)
+        self.assertEqual(st, 200)
+        st, res = self.g.reload_doc(doc_id)
+        self.assertIsNone(res)
+        self.assertEqual(st, 200)
+
     def test_list_doc_users(self):
         name = str(time.time_ns())
         st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
@@ -547,7 +655,7 @@ class TestDocs(BaseTestPyGrister):
         st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
         self.assertIsInstance(doc_id, str)
         self.assertEqual(st, 200)
-        st, res = self.g.download_sqlite(name+'.sqlite', 
+        st, res = self.g.download_sqlite(Path(f'{name}.sqlite'), 
             nohistory=True, template=True, doc_id=doc_id, team_id=self.team_id)
         self.assertIsNone(res)
         self.assertEqual(st, 200)
@@ -557,7 +665,7 @@ class TestDocs(BaseTestPyGrister):
         st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
         self.assertIsInstance(doc_id, str)
         self.assertEqual(st, 200)
-        st, res = self.g.download_excel(name+'.xls', table_id='Table1',
+        st, res = self.g.download_excel(Path(f'{name}.xls'), table_id='Table1',
                                         doc_id=doc_id, team_id=self.team_id)
         self.assertIsNone(res)
         self.assertEqual(st, 200)
@@ -567,7 +675,7 @@ class TestDocs(BaseTestPyGrister):
         st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
         self.assertIsInstance(doc_id, str)
         self.assertEqual(st, 200)
-        st, res = self.g.download_csv(name+'.csv', table_id='Table1',
+        st, res = self.g.download_csv(Path(f'{name}.csv'), table_id='Table1',
                                       doc_id=doc_id, team_id=self.team_id)
         self.assertIsNone(res)
         self.assertEqual(st, 200)
@@ -581,7 +689,7 @@ class TestDocs(BaseTestPyGrister):
                                           team_id=self.team_id)
         self.assertIsInstance(res, dict)
         self.assertEqual(st, 200)
-        st, res = self.g.download_schema('Table1', filename=name+'.json', 
+        st, res = self.g.download_schema('Table1', filename=Path(f'{name}.json'), 
                                          doc_id=doc_id, team_id=self.team_id)
         self.assertIsNone(res)
         self.assertEqual(st, 200)
@@ -601,7 +709,7 @@ class TestRecordAccess(BaseTestPyGrister):
         gristapi = api.GristApi(config=TEST_CONFIGURATION)
         st, res = gristapi.add_tables(tables, cls.doc_id, cls.team_id)
         cls.table_id = res[0]
-        total_apicalls.append(gristapi.apicalls)
+        total_apicalls.append(gristapi.apicaller.apicalls)
     
     def test_list_records(self):
         records = [{'Astr': 'test 1', 'Bnum': 1.1, 'Cint': 1, 'Dbol': True},
@@ -718,7 +826,7 @@ class TestConverters(BaseTestPyGrister):
         gristapi = api.GristApi(config=TEST_CONFIGURATION)
         st, res = gristapi.add_tables(tables, cls.doc_id, cls.team_id)
         cls.table_id = res[0]
-        total_apicalls.append(gristapi.apicalls)
+        total_apicalls.append(gristapi.apicaller.apicalls)
 
     def test_list_records(self):
         records = [{'Test': 'list_records', 'Sort': 'a', 
@@ -867,8 +975,8 @@ class TestTables(BaseTestPyGrister):
     @classmethod
     def setUpClass(cls):
         cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']
-        ws_id, name = _make_ws(cls.team_id)
-        cls.doc_id = _make_doc(ws_id)
+        cls.ws_id, name = _make_ws(cls.team_id)
+        cls.doc_id = _make_doc(cls.ws_id)
 
     def test_list_tables(self):
         st, res = self.g.list_tables(self.doc_id, self.team_id)
@@ -978,8 +1086,8 @@ class TestAttachments(BaseTestPyGrister):
     @classmethod
     def setUpClass(cls):
         cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']
-        ws_id, name = _make_ws(cls.team_id)
-        cls.doc_id = _make_doc(ws_id)
+        cls.ws_id, name = _make_ws(cls.team_id)
+        cls.doc_id = _make_doc(cls.ws_id)
 
     def test_list_attachments(self):
         st, res = self.g.list_attachments(doc_id=self.doc_id, team_id=self.team_id)
@@ -994,22 +1102,52 @@ class TestAttachments(BaseTestPyGrister):
         self.assertEqual(st, 200)
 
     def test_upload_download_attachments(self):
-        f = os.path.join(HERE, 'imgtest.jpg')
-        st, res = self.g.upload_attachment(f, doc_id=self.doc_id, 
-                                           team_id=self.team_id)
+        f = HERE / 'imgtest.jpg'
+        st, res = self.g.upload_attachments([f], doc_id=self.doc_id, 
+                                            team_id=self.team_id)
         self.assertIsInstance(res, list)
+        self.assertEqual(len(res), 1)
         self.assertEqual(st, 200)
-        name = 'att_'+str(time.time_ns())+'.jpg'
+        # upload multiple attachments
+        st, res = self.g.upload_attachments([f, f, f], doc_id=self.doc_id, 
+                                            team_id=self.team_id)
+        self.assertIsInstance(res, list)
+        self.assertEqual(len(res), 3)
+        self.assertEqual(st, 200)
+        name = Path(f'att_{time.time_ns()}.jpg')
         st, res = self.g.download_attachment(name, 1, doc_id=self.doc_id, 
                                              team_id=self.team_id)
         self.assertIsNone(res)
         self.assertEqual(st, 200)
+    
+    def test_download_restore_attachments(self):
+        # this is for the download/upload "archive" apis
+        f = HERE / 'imgtest.jpg'
+        st, res = self.g.upload_attachments([f], doc_id=self.doc_id, 
+                                            team_id=self.team_id)
+        self.assertEqual(st, 200)
+        name = Path(f'downloaded_atts_{time.time_ns()}.tar')
+        st, res = self.g.download_attachments(name, doc_id=self.doc_id, 
+                                              team_id=self.team_id)
+        self.assertEqual(st, 200)
+        self.assertIsNone(res)
+        self.assertTrue(name.is_file())
+        # now we re-upload the tar file
+        st, res = self.g.upload_restore_attachments(name, doc_id=self.doc_id, 
+                                                    team_id=self.team_id)
+        self.assertEqual(st, 200)
+        self.assertDictEqual(res, {'added': 0, 'errored': 0, 'unused': 1})
+        # downloading should work even if no attachments are present
+        doc_id = _make_doc(self.ws_id)
+        st, res = self.g.download_attachments(doc_id=doc_id, 
+                                              team_id=self.team_id)
+        self.assertEqual(st, 200)
 
     def test_see_attachment(self):
         # make sure we have at least one attachment to see
-        f = os.path.join(HERE, 'imgtest.jpg')
-        st, res = self.g.upload_attachment(f, doc_id=self.doc_id, 
-                                           team_id=self.team_id)
+        f = HERE / 'imgtest.jpg'
+        st, res = self.g.upload_attachments([f], doc_id=self.doc_id, 
+                                            team_id=self.team_id)
         self.assertIsInstance(res, list)
         self.assertEqual(st, 200)
         # note that id=1 could not be the one we just uploaded... 
@@ -1018,6 +1156,56 @@ class TestAttachments(BaseTestPyGrister):
         self.assertIsInstance(res, dict)
         self.assertEqual(st, 200)
 
+# to test external attachments, you must run a self-hosted instance of Grist 
+# with "GRIST_EXTERNAL_ATTACHMENTS_MODE=snapshots", and a S3-compatible bucket 
+@unittest.skipIf(TEST_CONFIGURATION['GRIST_TEST_RUN_EXT_ATTACH_TESTS'] == 'N', '')
+class TestExternalAttachments(BaseTestPyGrister):
+    @classmethod
+    def setUpClass(cls):
+        cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']
+        ws_id, name = _make_ws(cls.team_id)
+        cls.doc_id = _make_doc(ws_id)
+
+    def test_store(self):
+        # this is for the various "store" apis (see, update, settings)
+        st, res = self.g.see_attachment_store(doc_id=self.doc_id, 
+                                              team_id=self.team_id)
+        self.assertEqual(res, 'internal')
+        st, res = self.g.update_attachment_store(internal_store=False, 
+                            doc_id=self.doc_id, team_id=self.team_id)
+        self.assertIsNone(res)
+        st, res = self.g.see_attachment_store(doc_id=self.doc_id, 
+                                              team_id=self.team_id)
+        self.assertEqual(res, 'external')
+        st, res = self.g.list_store_settings(doc_id=self.doc_id, 
+                                             team_id=self.team_id)
+        self.assertEqual(st, 200) # not really much to verify here... 
+
+    def test_transfer(self):
+        # this is for the transfering files apis, between external/internal store
+        st, res = self.g.update_attachment_store(doc_id=self.doc_id, 
+                                                 team_id=self.team_id)
+        st, res = self.g.see_attachment_store(doc_id=self.doc_id, 
+                                              team_id=self.team_id)
+        self.assertEqual(res, 'internal')
+        f = HERE / 'imgtest.jpg'
+        st, res = self.g.upload_attachments([f, f, f], doc_id=self.doc_id, 
+                                            team_id=self.team_id)
+        st, res = self.g.update_attachment_store(internal_store=False, 
+                            doc_id=self.doc_id, team_id=self.team_id)
+        st, res = self.g.see_attachment_store(doc_id=self.doc_id, 
+                                              team_id=self.team_id)
+        self.assertEqual(res, 'external')
+        # now we should have some attachments to transfer
+        st, res = self.g.transfer_attachments(doc_id=self.doc_id, 
+                                              team_id=self.team_id)
+        self.assertTrue(res['status']['isRunning'])
+        # "see_transfer_status" return depends on when we call it... 
+        st, res = self.g.see_transfer_status(doc_id=self.doc_id, 
+                                             team_id=self.team_id)
+        self.assertEqual(st, 200) # little we can test
+
+
 class TestWebhooks(BaseTestPyGrister):
     @classmethod
     def setUpClass(cls):
